#include "display.h"
#include <string.h>

struct display_values display;

// loads size bytes from data to position cgaddr in the displays cgram.
// expects characters to be 8 lines high
void load_cgram(display_t id, position_t cgaddr, cgram_t data, int size)
{
    position_t line;
    position_t chr;
#ifdef TEST
    printf("cgram(%02x, %04x)\n", id, cgaddr);
    for (chr=0 ; chr<size/8 ; chr++) {
        printf(" %8d", chr);
    }
    printf("\n");
    for (line=0 ; line<8 ; line++) {
        for (chr=0 ; chr<size/8 ; chr++) {
            printf("%c%c%c%c%c%c",
                    data[chr][line]&0xe0?'!':' ',
                    data[chr][line]&0x10?'*':'.',
                    data[chr][line]&0x08?'*':'.',
                    data[chr][line]&0x04?'*':'.',
                    data[chr][line]&0x02?'*':'.',
                    data[chr][line]&0x01?'*':'.');
        }
        printf("\n");
    }
#endif
    dispinsn(id, 0x40+(cgaddr&0x3f));
    for (chr=0 ; chr<size/8 ; chr++) {
        for (line=0 ; line<8 ; line++) {
            dispdata(id, data[chr][line]);
        }
    }
}

// copies <size> characters to ddram position [line,col] from <data>
void load_ddram(display_t id, position_t line, position_t col, ddramline_t data, int size)
{
    position_t chr;
    if (size<0)
        size=strlen(data);
#ifdef TEST
    printf("display: [%d,%d]\n", line, col);
    for (chr=0 ; chr<size ; chr++) {
        printf(" %02x", data[chr]);
    }
    printf("\n");
#endif
    dispinsn(id, 0x80+0x40*line+col);
    for (chr=0 ; chr<size ; chr++) {
        dispdata(id, data[chr]);
    }
}

// writes <size> spaces to ddram position [line,col]
void erase_chars(display_t id, position_t line, position_t col, int n)
{
    dispinsn(id, 0x80+0x40*line+col);
    while(n--)
        dispdata(id, ' ');
}


void voter_writestring(char *str)
{
    int id= 8+8*(display.currow/2);
    dispinsn(id, 0x80+0x40*(display.currow%2)+display.curcol);
    while (*str) {
        if (*str=='\n') {
            display.curcol=40;
        }
        if (*str=='\t') {
            display.curcol = 5*(display.curcol/5+1);
        }
        else {
            dispdata(id, *str);
            display.curcol++;
        }

        if (display.curcol>=40) {
            display.curcol= 0;
            display.currow++;
            if (display.currow==4) {
                display.currow=0;
            }
            id= 8+8*(display.currow/2);
            dispinsn(id, 0x80+0x40*(display.currow%2)+display.curcol);
        }
        str++;
    }
}

