#ifndef __DISPLAY_H__
#define __DISPLAY_H__

#ifdef TEST
#include <stdio.h>
#endif

#include "io_funcs.h"

typedef char cgramchar_t;
typedef cgramchar_t charbitmap_t[8];
typedef charbitmap_t cgram_t[8];
typedef char ddramchar_t;
typedef ddramchar_t ddramline_t[64];
typedef ddramline_t ddram_t[2];

typedef short position_t;
typedef char  display_t;

enum display_instructions_t {
    disp_clearscreen=1,
    disp_homecursor=2,
    disp_entrymode=4,
    disp_onoff=8,
    disp_shift=16,
    disp_setparams=32,
    disp_cgram=64,
    disp_ddram=128,
};
/*
enum display_control_t {
    disp_read=1,
    disp_instruction=2,
    operator_display=4,
    votertop_display=8,
    voterbottom_display=0x10,
    disp_backlight=0x20,
};
*/
struct display_values {
    char backlight_value;
    int currow;
    int curcol;
};
extern struct display_values display;

#ifndef TEST

// controls the various chipselect/register/rw line
static inline void dispcontrol(char val) {
    *(volatile unsigned char*)0xBFFFFB = display.backlight_value|val;
}
// write data to the display
static inline void set_dispdata(char val) {
    *(volatile unsigned char*)0xBFFFF9 = val;
}
// read data from the display
static inline char dispstatus() {
    return *(volatile unsigned char*)0xBFFFF8;
}
#endif

// reads the display status byte
// bit7 = busy
// bit6-0 = current address
static inline char read_dispstatus(display_t id)
{
#ifndef TEST
    char status;
    dispcontrol(1);
    dispcontrol(1|id);
    status= dispstatus();
    dispcontrol(1);

    return status;
#else
    return 0;
#endif
}

// waits while display is busy
static inline void wait_for_display(display_t id)
{
    while (read_dispstatus(id)&0x80)
        watchdog();
}

// send an instruction to the display
static inline void dispinsn(display_t id, int instruction)
{
#ifndef TEST
    wait_for_display(id);
    set_dispdata(instruction);
    dispcontrol(0);
    dispcontrol(0|id);
    dispcontrol(0);
    dispcontrol(3);
#else
    printf("%02x: instr(0x%02x);\n", id, instruction);
#endif
}

// sends data to the display
static inline void dispdata(display_t id, int data)
{
#ifndef TEST
    wait_for_display(id);
    set_dispdata(data);
    dispcontrol(2);
    dispcontrol(2|id);
    dispcontrol(2);
    dispcontrol(3);
#else
    printf("%02x: data(0x%02x);\n", id, data);
#endif
}

// turns on the voter display backlight
static inline void backlight(char on)
{
    display.backlight_value= on?0x20:0;
#ifndef TEST
    dispcontrol(3);
#endif
}

// initializes and clears the display
static inline void init_display(display_t id)
{
    dispinsn(id, 0x38); // 8bit, 8*5 font, 2 lines
    dispinsn(id, 0x0c); // disp on, no cursor, no blink
    dispinsn(id, 0x06); // set cursor mode, increment, scroll off
    dispinsn(id, 0x01); // clear screen
}

void voter_writestring(char *str);
void load_cgram(display_t id, position_t cgaddr, cgram_t data, int size);
void load_ddram(display_t id, position_t line, position_t col, ddramline_t data, int size);
void erase_chars(display_t id, position_t line, position_t col, int n);
#endif
