#include <string.h>
#include "eeprom.h"
#include "io_funcs.h"   // for watchdog

void eeprom_read_data(long dwOffset, char *data, int len)
{
    volatile char *ptr= (char*)dwOffset;
    while (len--) {
        *data++ = *ptr;

        ptr += 2;
    }
}
int eeprom_wait(volatile char *ptr)
{
    int i=0;
    // the 28HC64D toggles bit6 of the databus, while it is busy programming a byte.
    while (*ptr!=*ptr) {
        watchdog();
        i++;
    }
    return i;
}
void eeprom_write_byte(volatile char *eeptr, char data)
{
    *eeptr = data;

    // wait until device is ready when writing the last address in a page
    if (((long)eeptr&0x7e)==0x7e)
        eeprom_wait(eeptr);
}
int eeprom_write_data(long dwOffset, char *data, int len)
{
    volatile char *ptr= (char*)dwOffset;

    if (len==-1)
        len= strlen(data)+1;
    while (len) {
        eeprom_write_byte(ptr, *data++);

        ptr += 2;
        len--;
    }
    return eeprom_wait(ptr);
}
short eeprom_read_value(long dwOffset)
{
    volatile char *ptr= (char*)dwOffset;
    return ptr[0]*256 + ptr[2];
}
int eeprom_write_value(long dwOffset, short value)
{
    volatile char *ptr= (char*)dwOffset;
    eeprom_write_byte(ptr, value>>8);
    ptr += 2;
    eeprom_write_byte(ptr, value&0xff);

    return eeprom_wait(ptr);
}

void eeprom_erase(long offset, int len)
{
    char buf[256];
    memset(buf, 0, 256);
    while (len) {
        int want= len;
        if (want > 256)
            want= 256;
        eeprom_write_data(offset, buf, want);
        offset+=want;
        len -=want;
    }
}
