#define VRIJGAVE2          0x04
#define VRIJGAVE1          0x08
#define STEM_SLEUTEL       0x10
#define PROGRAMEER_SLEUTEL 0x20
#define FUNCTIE_SLEUTEL    0x40
#define SERVICE_SWITCH     0x80

extern char bc000c;

const unsigned char NUMBER_OF_OUTLINES=40;
const unsigned char NUMBER_OF_INLINES=5;
unsigned char * const keyboard_base = (unsigned char *)0xbc0000;

unsigned char g_ol, g_il, g_bit;

unsigned char number_of_bits_in_nibble[16] 
  = { 0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4 };

unsigned char leading_bit_l[16]
  = { 0, 0, 1, 255, 2, 255, 255, 255, 3, 255, 255, 255, 255, 255, 255, 255 };
unsigned char leading_bit_h[16]
  = { 0, 4, 5, 255, 6, 255, 255, 255, 7, 255, 255, 255, 255, 255, 255, 255 };

unsigned char x_mapping[]
  = { 0x24, // does not happen on this pad
      0x08, 0x07, 0x06, 0x05, 0x04, 0x03, 0x02, 0x01, 0x00,
      0x25, // does not happen on this pad
      0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11,
      0x26, // does not happen on this pad
      0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20, 0x21, 0x22, 0x23,
      0x27, // does not happen onb this pad
      0x1a, 0x19, 0x18, 0x17, 0x16, 0x15, 0x14, 0x13, 0x12,
};

void set_out_line(unsigned char index) {
  keyboard_base[1] = keyboard_base[3] = keyboard_base[5] = keyboard_base[7] = keyboard_base[9] = 0;
  keyboard_base[(index>>2)|1] = ((unsigned char)1) << (index&7);
}

inline unsigned char number_of_bits_in_byte (unsigned char c) {
  return number_of_bits_in_nibble[c & 0x0f] + number_of_bits_in_nibble[(c>>4) & 0x0f];
}

inline unsigned char find_leading_bit (unsigned char c) {
  return ( number_of_bits_in_byte(c) == 1 ) ? leading_bit_l[c & 0xf] | leading_bit_h[(c>>4) & 0xf] : 0xff;
}

long read_from_keyboard( ) {
  unsigned char iline, oline;
  int key_code=-2;

  for( oline=0; oline<NUMBER_OF_OUTLINES; ++oline) {
    set_out_line(oline);
    for(iline = 0; iline < NUMBER_OF_INLINES; ++iline) {
      unsigned char top_bit, bit_mask = keyboard_base[iline<<1];

      if( bit_mask) {
        top_bit = find_leading_bit(bit_mask);
        if( (key_code >=0) || (top_bit == 0xff)) return -1;

        // x-axis - correction .. we know of 4 areas, 9 lines each
        if( oline <= 0x1e )
          key_code = x_mapping[iline*8+top_bit];
        else
          key_code = iline*8+top_bit;

        key_code <<= 8;

        // y-axis - correction for the lower left area
        if( (iline <= 2) && (oline>0x0e))
          key_code += ( 0x2d - oline );
        else
          key_code += oline;

        // debug
        g_il=iline; g_ol=oline; g_bit=top_bit;
      }
    }
  }
  return key_code;
}


