#include "memmodule.h"
#include "io_funcs.h"   // delay, watchdog
#include <string.h>

void vpp_enable_even()
{
    *(volatile char*)0xbffff5 = 1;
}
void vpp_enable_odd()
{
    *(volatile char*)0xbffff5 = 3;
    *(volatile char*)0xbffff5 = 2;
    delay(500);
    *(volatile char*)0xbffff5 = 6;
    delay(4);
    watchdog();
}

void vpp_disable()
{
    *(volatile char*)0xbffff5 = 0;
}
char *get_mm_base(mmid_t id, mm_select_t part)
{
    return (char*)(id<<16)+(part-1);
}

#define MM_INT8  -1
#define MM_7INT4 -2
struct paraminfo_t {
    int offset;
    int type;   // <0 : specifies type, >0 : fixedlen string, =0 NUL term string
} paraminfo[]= {
    { 0x00, MM_INT8 },  // MM_TYPE,                //  i Module type  'N', '2', '1'
    { 0x02, MM_INT8 },  // MM_IDENTIFIER,          //  i A1 or A3 - identifies odd/even flash chip A2 or A4 - identifies odd/even flash chip
    { 0x04, MM_7INT4 }, // MM_ERASECOUNT,          //  i Erasure count (stored as 7 bits per byte)
    { 0x08,  8 },       // MM_MODULEID,            //  s Module id
    { 0x18,  8 },       // MM_VOTINGDATE,          //  s Voting date
    { 0x28, MM_INT8 },  // MM_OPENVOTE,            //  i FF = (1st?) election open, 00 = election closed) 
    { 0x2c, MM_7INT4 }, // MM_CHECKSUM,            //  i Checksum of candidate list, stored as 7bits-per-byte, 4 bytes, but only 28 bits are used.
    { 0x30,  8 },       // MM_COMPUTERID,          //  s Voting computer serial number, every byte repeated once, docs say recorded on first vote 
    { 0x40,  0 },       // MM_VOTE_DESCRIPTION,    //  s Null terminated string with name of the election
    { 0x90, MM_INT8 },  // MM_VOTE_TYPE,           //  i Two bytes 31 31  - election type
};
#define MAX_PARAMINFO (sizeof(paraminfo)/sizeof(struct paraminfo_t))
unsigned long mm_read_intparam(mmid_t id, mm_select_t part, mm_param_t param)
{
    if (param<0 || param>=MAX_PARAMINFO)
        return 0;
    volatile unsigned char *ptr= get_mm_base(id, part==MM_BOTH?MM_EVEN:part);

    ptr += paraminfo[param].offset;

    switch(paraminfo[param].type) {
        case MM_INT8:
            return ptr[0]&0x7f;
        case MM_7INT4:
            if (part==MM_BOTH)
                return ((((((ptr[0]&0x7f)<<7)|(ptr[1]&0x7f))<<7)|(ptr[2]&0x7f))<<7)|(ptr[3]&0x7f);
            else if (part==MM_EVEN)
                return ptr[0]*256+ptr[2];
            else if (part==MM_ODD)
                return ptr[1]*256+ptr[3];
    }
    return 0;
}
void mm_write_intparam(mmid_t id, mm_select_t part, mm_param_t param, long value)
{
}
void mm_read_stringparam(mmid_t id, mm_select_t part, mm_param_t param, char *str, int len)
{
    if (param<0 || param>=MAX_PARAMINFO)
        return;
    volatile char *ptr= get_mm_base(id, part==MM_BOTH?MM_EVEN:part);

    ptr += paraminfo[param].offset;

    if (paraminfo[param].type<0) {
        return;
    }
    else if (paraminfo[param].type==0) {
        int i;
        for (i=0 ; i<len ; i++)
            str[i]= ptr[i]&0x7f;
        str[len-1]= 0;
    }
    else {
        int i;
        if (part==MM_BOTH) {
            for (i=0 ; i<2*paraminfo[param].type ; i++)
                str[i] = ptr[i];
        }
        else {
            for (i=0 ; i<paraminfo[param].type ; i++)
                str[i]= ptr[i*2 +( part==MM_EVEN?0 : 1)]&0x7f;
            str[i]= 0;
        }
    }
    return;
}
void mm_write_stringparam(mmid_t id, mm_select_t part, mm_param_t param, char *str, int len)
{
}

int mm_verify_erase(mmid_t id, mm_select_t part, int offset)
{
    volatile char *ptr= get_mm_base(id, part==MM_BOTH?MM_EVEN:part);
    volatile char *p;
    *ptr=0xA0;
    delay(1000);
    for (p=ptr ; p<ptr+0x40000 ; p+=2)
        if (*p!=(char)0xff)
            return 0;
    return 1;
}


int mm_erase(mmid_t id, mm_select_t part)
{
    volatile char *ptr= get_mm_base(id, part);

    int tries=0;
    while (1) {
        // do erase
        *ptr=0x80;
        *ptr=0x80;

        delay(1000);

        if (tries>1000)
            return 0;
    }
    return 1;
}

int mm_write_byte(mmid_t id, mm_select_t part, int offset, char data)
{
    int i;
    if (part==MM_BOTH)  // 'both' not allowed
        return 0;
    volatile char *ptr= get_mm_base(id, part)+offset;

    watchdog();
    for (i=0 ; i<25 ; i++) {
        *ptr= 0x40;
        *ptr= data;
        delay(4);
        *ptr= 0x60;
        delay(3);
        if (*ptr==data)
            return 1;
    }
    return 0;
}

void mm_write(mmid_t id, mm_select_t part, int offset, char*data, int size)
{
    if (part==MM_BOTH) {
        mm_write(id, MM_EVEN, offset, data, size);
        mm_write(id, MM_ODD, offset+1, data+1, size-1);
    }
    else {
        while (size>0) {
            mm_write_byte(id, part, offset, *data);
            size-=2;
            offset+=2;
            data+=2;
        }
    }
}

int flash_write_byte(int offset, char data)
{
    int i;
    volatile char *ptr= (volatile char*)offset;

    watchdog();
    for (i=0 ; i<25 ; i++) {
        *ptr= 0x40;
        *ptr= data;
        delay(4);
        *ptr= 0x60;
        delay(3);
        if (*ptr==data)
            return 1;
    }
    return 0;
}
void flash_write_data(int offset, char *data, int size)
{
    vpp_enable_even();
    tsleep(35);
    if (offset&1)
        vpp_enable_odd();

    while (size) {
        flash_write_byte(offset, *data);
        data++;
        size--;
        offset+=2;
    }
    vpp_disable();
}
