#ifndef __MMODULE_H__
#define __MMODULE_H__

typedef enum {
    MM_TYPE,                // 0000 i Module type  'N', '2', '1'
    MM_IDENTIFIER,          // 0002 i A1 or A3 - identifies odd/even flash chip A2 or A4 - identifies odd/even flash chip
    MM_ERASECOUNT,          // 0004 i Erasure count (stored as 7 bits per byte)
    MM_MODULEID,            // 0008 s Module id
    MM_VOTINGDATE,          // 0018 s Voting date
    MM_OPENVOTE,            // 0028 i FF = (1st?) election open, 00 = election closed) 
    MM_CHECKSUM,            // 002c i Checksum of candidate list, stored as 7bits-per-byte, 4 bytes, but only 28 bits are used.
    MM_COMPUTERID,          // 0030 s Voting computer serial number, every byte repeated once, docs say recorded on first vote 
    MM_VOTE_DESCRIPTION,    // 0040 s Null terminated string with name of the election
    MM_VOTE_TYPE,           // 0090 i Two bytes 31 31  - election type
} mm_param_t;

typedef enum {
    MM_PROG1=0x3c,
    MM_PROG2=0x3e,
    MM_VOTE1=0x7c,
    MM_VOTE2=0x7e,
} mmid_t;
typedef enum {
    MM_EVEN=1,
    MM_ODD=2,
    MM_BOTH=3,
} mm_select_t;

unsigned long mm_read_intparam(mmid_t id, mm_select_t part, mm_param_t param);
void mm_write_intparam(mmid_t id, mm_select_t part, mm_param_t param, long value);
void mm_read_stringparam(mmid_t id, mm_select_t part, mm_param_t param, char *str, int len);
void mm_write_stringparam(mmid_t id, mm_select_t part, mm_param_t param, char *str, int len);

int mm_erase(mmid_t id, mm_select_t part);
void mm_write(mmid_t id, mm_select_t part, int offset, char*data, int size);

void flash_write_data(int offset, char *data, int size);

/*
0092-1000 .. also seems to have significance
1000 Candidate lists per party.
   36 party slots with 30 candidates each.
   each party list is 155 bytes long, with 5 bytes per candidate
   the first entry points to the party name
   format
      +00: (1 byte) List Number
      +01: (1 byte) Candidate number (0=party name )
      +02: (3 bytes) (7 bits-per-byte): Offset from 0x1000 to string
25cc-28a4 Null terminated strings, first five party names, then all candidate names
28a5-13fff Random data (see below)
13000-14000 .. also seems to have significance
14000-1fffff Single votes
  14000-19ffff Normal storage
  1a000-1fffff Inverted storage
*/

#endif

