#include "uart.h"
#include "io_funcs.h"
#include <stdarg.h>
#include <stdio.h>

/*
 *  dev5 - com1 - modem  - bfffff.bit1 - uartinfo[0] - 00FC0001, 00FC0003
 *  dev6 - com2 - pc     - bfffff.bit0 - uartinfo[1] - 00FC0005, 00FC0007
 *
 *  com1 is closest to the edge of the circuit board
 *  com2 is the one also used in the rw unit
 */
/*
struct uart_info {
    char buffer[256];
    char rdptr;
    char wrptr;
    char bufsize;
};

struct uart_info uartinfo;
*/
#define UARTCONTROL(id) (*(volatile char*)(0xfc0003+4*id))
#define UARTDATA(id) (*(volatile char*)(0xfc0001+4*id))

void init_uart(uartid_t id)
{
    bfffff.uart_value=2-id; // high baudrate
    update_bfffff();

    UARTCONTROL(id)=0x55;   // 9600, 8n1, noirq
    /*
    uartinfo.rdptr=0;
    uartinfo.wrptr=0;
    uartinfo.bufsize=0;
    */
}

char read_uart(uartid_t id/*, int timeout*/)
{
    while (!(UARTCONTROL(id)&1))
        watchdog();

    return UARTDATA(id);
}
int read_uart_noblock(uartid_t id, char *pchr)
{
    if (!(UARTCONTROL(id)&1))
        return -1;

    *pchr = UARTDATA(id);

    return 0;
}
void readline_uart(uartid_t id, char *line, int size)
{
    char c;
    while (size--) {
        c = read_uart(id);
        if (c=='\n' || c=='\r')
            break;
        *line++ = c;
    }
    *line = 0;
}

void write_uart(uartid_t id, char c)
{
    while (!(UARTCONTROL(id)&2))
        watchdog();

    UARTDATA(id)= c;
}


void writestring_uart(uartid_t id, char *s)
{
    while (*s)
        write_uart(id, *s++);
}

void uprintf(char *fmt, ...)
{
    char msg[255];
    va_list ap;

    va_start(ap, fmt);
    vsprintf(msg, fmt, ap);
    va_end(ap);
    writestring_uart(1, msg);
}

