#include <stdio.h>
#include "uart.h"
#include <stdlib.h>

#include <stdarg.h>
#include <string.h>
#include "misc.h"
#include "display.h"
#include "io_funcs.h"
#include "eeprom.h"
#include "memmodule.h"

void eeprom_dump(long dwOffset, long len)
{
    int i;
    char ascii[17];
    volatile char *ptr= (char*)dwOffset;
    char sbuf[255];
    int si= 0;

    memset(ascii, ' ', 16);
    ascii[16]= 0;

    if (dwOffset&0x1f) {
        si+=sprintf(sbuf+si, "%08lx: ", dwOffset&~0x1f);
        for (i=0 ; i<(dwOffset&0x1f)/2 ; i++) {
            si+=sprintf(sbuf+si, "   ");
        }
    }
    for (i=0 ; i<len ; i++) {
        if ((i&0xf)==0) {
            si+=sprintf(sbuf+si, "%08lx: ", dwOffset+2*i);
        }
        si+=sprintf(sbuf+si, " %02x", (unsigned char)*ptr);
        ascii[i&15] = (*ptr<' ' || *ptr >'~') ? '.' : *ptr;

        if ((i&0xf)==0xf) {
            si+=sprintf(sbuf+si, " - %s\r\n", ascii);
            writestring_uart(1, sbuf);
            si=0;
            char c;
            if (0==read_uart_noblock(1, &c) && c==3)
                return;
        }
        ptr += 2;
    }
    if (i&0xf) {
        i &= 0xf;
        for ( ; i < 16 ; i++) {
            si+=sprintf(sbuf+si, "   ");
            ascii[i]=' ';
        }
        si+=sprintf(sbuf+si, " - %s\r\n", ascii);
        writestring_uart(1, sbuf);
        si=0;
    }
}
void mem_dump(long dwOffset, long len)
{
    int i;
    char ascii[17];
    volatile char *ptr= (char*)dwOffset;
    char sbuf[255];
    int si= 0;

    memset(ascii, ' ', 16);
    ascii[16]= 0;

    if (dwOffset&0xf) {
        si+=sprintf(sbuf+si, "%08lx: ", dwOffset&~0xf);
        for (i=0 ; i<(dwOffset&15) ; i++) {
            si+=sprintf(sbuf+si, "   ");
        }
    }
    for (i=0 ; i<len ; i++) {
        unsigned char c= *ptr;
        if ((i&0xf)==0) {
            si+=sprintf(sbuf+si, "%08lx: ", dwOffset+i);
        }
        si+=sprintf(sbuf+si, " %02x", c);
        c &= 0x7f;
        ascii[i&15] = (c<' ' || c >'~') ? '.' : c;

        if ((i&0xf)==0xf) {
            si+=sprintf(sbuf+si, " - %s\r\n", ascii);
            writestring_uart(1, sbuf);
            si=0;
            char c;
            if (0==read_uart_noblock(1, &c) && c==3)
                return;
        }
        ptr ++;
    }
    if (i&0xf) {
        i &= 0xf;
        for ( ; i < 16 ; i++) {
            si+=sprintf(sbuf+si, "   ");
            ascii[i]=' ';
        }
        si+=sprintf(sbuf+si, " - %s\r\n", ascii);
        writestring_uart(1, sbuf);
        si=0;
    }
}
void usage()
{
    uprintf("dump offset length\r\n");
    uprintf("\r\n");
    uprintf("eeprom commands:\r\n");
    uprintf("   edump eoffset length\r\n");
    uprintf("   string eoffset maxlen\r\n");
    uprintf("   value eoffset\r\n");
    uprintf("   wstr eoffset string\r\n");
    uprintf("   wbyte eoffset byes\r\n");
    uprintf("      numbers may be decimal or 0xhex\r\n");
    uprintf("0x800000 is always added to eeprom offsets\r\n");
}

// using my own implementation, the newlib one is crashing
char *strtok_x(char *str, char *sep)
{
    static char *mystr;
    if (str)
        mystr= str;
    if (*mystr==0)
        return 0;

    char *start;
    while (*mystr==*sep && *mystr!=0)
        *mystr++= 0;
    start= mystr;
    while (*mystr!=*sep && *mystr!=0)
        mystr++;
    if (*mystr==*sep)
        *mystr++= 0;
    return start;
}
void dispprintf(char *fmt, ...)
{
    char msg[255];
    va_list ap;

    va_start(ap, fmt);
    vsprintf(msg, fmt, ap);
    va_end(ap);
    voter_writestring(msg);
}

void readwecho(char *line, int size)
{
    static char saved[10][80];
    static short isave=0;
    int i;
    char *p= line;
    while(1)
    {
        char c= read_uart(1);
        if (c=='\r' || c=='\n')
            break;
        if (c==0x8) {
            p--;
            if (p<line)
                p=line;
            write_uart(1,0x8);
            write_uart(1,0x20);
            write_uart(1,0x8);
        }
        else if (c==0x10) { // ctrl-p
            strncpy(saved[isave], line, 80);
            write_uart(1, 0xd);
            for (i=0 ; i<strlen(line) ; i++)
                write_uart(1,0x20);
            write_uart(1, 0xd);

            isave--; isave%=10;
            strncpy(line, saved[isave], size);
            uprintf("%s", line);

        }
        else if (c==0xe) {  // ctrl-n
            strncpy(saved[isave], line, 80);
            write_uart(1, 0xd);
            for (i=0 ; i<strlen(line) ; i++)
                write_uart(1,0x20);
            write_uart(1, 0xd);
            isave++; isave%=10;
            strncpy(line, saved[isave], size);
            uprintf("%s", line);
        }
        else {
            if (p<line+size-1) {
                *p++ = c;
                *p=0;
                write_uart(1,c);
            }
        }
    }
    *p=0;
    strncpy(saved[isave++], line, 80);
    isave%=10;
}

char *eeprom_read_string(long offset, char *buf, int len)
{
    eeprom_read_data(offset, buf, len);
    buf[len]= 0;
    return buf;
}

#ifdef TEST
void entrypoint()
#else
int main()
#endif
{
    init_display(4); load_ddram(4,0,0, "init04",-1);
    wbeep(1000);
    backlight(1);

    watchdog();
/*
    for (i=0 ; i<1000 ; i++) {
        int w;
        char str[16];
        dispprintf("%08lx: %04x : %5d\n", timer_value, (unsigned short)eeprom_read_value(0x803FD0), eeprom_write_value(0x803FD0, timer_value));
        w= eeprom_write_data(0x803fc0, "test1234567", 12);
        eeprom_read_data(0x803fc0, str, 16);
        dispprintf("%08lx: %5d %s\n", timer_value, w, str);
        tsleep(10);
    }
    {
        char buf[16];
        memset(buf,0,16);
        eeprom_write_data(0x803fc0, buf, 16);
    }
*/

    load_ddram(4,0,10, "uart#1",-1);
    init_uart(1); uprintf("uart11111111111111\r\n");
    load_ddram(4,1,0, "commandloop",-1);
    while (1) {
        char line[80];
        long params[4];
        char *argv[4];
        int argc;
        uprintf("#");
        readwecho(line, 80);
        uprintf("\r\n");
        load_ddram(4,0,0, line, -1);
        argc=0;
        argv[argc] = strtok_x(line, " ");
        params[argc] = 0;
        if (argv[argc]) {
            while (++argc<4) {
                argv[argc] = strtok_x(NULL, " ");
                if (argv[argc]==NULL)
                    break;
                params[argc] = strtoul(argv[argc],0,0);
            }
        }
        if (argv[0]==NULL)
            continue;

        //uprintf("%d args: %s=%08lx %s=%08lx %s=%08lx\r\n", argc, argv[0], params[0], argv[1], params[1], argv[2], params[2]);
        
        if (strcmp(argv[0], "edump")==0) {
            if (argc==3) {
                eeprom_dump(0x800000+params[1], params[2]);
            }
            else {
                uprintf("edump needs 2 args\r\n");
            }
        }
        else if (strcmp(argv[0], "dump")==0) {
            if (argc==3) {
                mem_dump(params[1], params[2]);
            }
            else {
                uprintf("dump needs 2 args\r\n");
            }
        }
        else if (strcmp(argv[0], "step")==0) {
            if (argc==4) {
                int ofs;
                for (ofs=params[1] ; ofs<params[1]+params[2] ; ofs+=params[3])
                    mem_dump(ofs, 16);
            }
            else {
                uprintf("step needs 3 args\r\n");
            }
        }
        else if (strcmp(argv[0], "string")==0) {
            char strbuf[255];
            if (argc==3) {
                eeprom_read_data(0x800000+params[1], strbuf, params[2]); strbuf[254]=0;
                uprintf("read string from %08lx maxl=%ld\r\n", params[1], params[2]);
                uprintf("  %s\r\n", strbuf);
            }
            else {
                uprintf("string needs 2 args\r\n");
            }
        }
        else if (strcmp(argv[0], "value")==0) {
            if (argc==2) {
                short value=eeprom_read_value(0x800000+params[1]);
                uprintf("read value from %08lx : %04x\r\n", params[1], value);
            }
            else {
                uprintf("value needs 1 arg\r\n");
            }
        }
        else if (strcmp(argv[0], "wstr")==0) {
            if (argc==3) {
                eeprom_write_data(0x800000+params[1], argv[2], strlen(argv[2]));
            }
            else {
                uprintf("wstr needs 2 args\r\n");
            }
        }
        else if (strcmp(argv[0], "wbyte")==0) {
            if (argc==3) {
                char byte= params[2];
                eeprom_write_data(0x800000+params[1], &byte, 1);
            }
            else {
                uprintf("wbyte needs 2 args\r\n");
            }
        }
        else if (strcmp(argv[0], "fstr")==0) {
            if (argc==3) {
                flash_write_data(params[1], argv[2], strlen(argv[2]));
            }
            else {
                uprintf("wbyte needs 2 args\r\n");
            }
        }
        else if (strcmp(argv[0], "erase")==0) {
            if (argc==3) {
                eeprom_erase(0x800000+params[1], params[2]);
            }
            else {
                uprintf("erase needs 2 args\r\n");
            }
        }
        else if (strcmp(argv[0], "eeinfo")==0) {
            volatile unsigned char *p;
            int evcnt=0;
            uprintf("\r\n");
            for (p = (volatile unsigned char *)0x800000 ; p<(volatile unsigned char *)0x802000 ; p+=8) {
                if (p[0]!=0xff || p[2]!=0xff || p[4]!=0xff || p[6]!=0xff) {
                    if (evcnt<32)
                        uprintf("%5d  %5d\r\n", p[0]+256*p[2], p[4]+256*p[6]);
                    else if (evcnt==32)
                        uprintf("more events not listed\r\n");
                    evcnt++;
                }
            }
            uprintf("total: %d events\r\n", evcnt);

            uprintf("\r\n");
            int ercnt= 0;
            for (p = (volatile unsigned char *)0x802000 ; p<(volatile unsigned char *)0x803da0; p+=8) {
                if (p[0]!=0xff || p[2]!=0xff || p[4]!=0xff || p[6]!=0xff) {
                    if (ercnt<32)
                        uprintf("%5d  %5d\r\n", p[0]+256*p[2], p[4]+256*p[6]);
                    else if (ercnt==32)
                        uprintf("more errors not listed\r\n");
                    ercnt++;
                }
            }
            uprintf("total: %d errors\r\n", ercnt);

            char sbuf[65];

            uprintf("00803FDC voter panel = %d\r\n", *(unsigned char*)0x803FDC);
            uprintf("00803FDE top panel   = %d\r\n", *(unsigned char*)0x803FDE);
            uprintf("00803FE0 deviceid    = %s\r\n", eeprom_read_string(0x803FE0, sbuf, 8));
            uprintf("00803FF0 has modem   = %d\r\n", *(unsigned char*)0x803FF0);
            uprintf("00803FF2 has blind   = %d\r\n", *(unsigned char*)0x803FF2);
            uprintf("00803FF4 devtype     = %d\r\n", *(unsigned char*)0x803FF4);
            uprintf("00803FF6 beeper      = %d\r\n", *(unsigned char*)0x803FF6);
            uprintf("00803FF8 scroll      = %d\r\n", *(unsigned char*)0x803FF8);
            uprintf("00803FFA modem baud  = %d\r\n", *(unsigned char*)0x803FFA);
            uprintf("00803FFC pc baud     = %d\r\n", *(unsigned char*)0x803FFC);

        }
        else if (strcmp(argv[0], "mminfo")==0) {
            int mid= (argc!=2) ? 0x7c : params[1];
            uprintf("%02x0000 type   %02x\r\n", mid, mm_read_intparam(mid, MM_EVEN, MM_TYPE));
            uprintf("%02x0002 id     %02x\r\n", mid, mm_read_intparam(mid, MM_EVEN, MM_IDENTIFIER));
            uprintf("%02x0004 erase  %08x\r\n", mid, mm_read_intparam(mid, MM_BOTH, MM_ERASECOUNT));
            uprintf("%02x0028 open   %02x\r\n", mid, mm_read_intparam(mid, MM_EVEN, MM_OPENVOTE));
            uprintf("%02x002c chksum %08x\r\n", mid, mm_read_intparam(mid, MM_BOTH, MM_CHECKSUM));
            uprintf("%02x0090 vote   %02x\r\n", mid, mm_read_intparam(mid, MM_EVEN, MM_VOTE_TYPE));
            char str[81];
            mm_read_stringparam(mid, MM_EVEN, MM_MODULEID, str, 9);          uprintf("%02x0008 modid  %s\r\n", mid, str);
            mm_read_stringparam(mid, MM_EVEN, MM_VOTINGDATE, str, 9);        uprintf("%02x0018 date   %s\r\n", mid, str);
            mm_read_stringparam(mid, MM_EVEN, MM_COMPUTERID, str, 9);        uprintf("%02x0030 compid %s\r\n", mid, str);
            mm_read_stringparam(mid, MM_BOTH, MM_VOTE_DESCRIPTION, str, 81); uprintf("%02x0040 desc   %s\r\n", mid, str);

            int cnt0092=0;
            volatile char *p;
            for (p=(volatile char *)0x7c0092 ; p<(volatile char *)0x7c1000 ; p+=5)
                if (((*p)&0x7f)!=0x7f)
                    cnt0092++;
            uprintf("0092 list contains %d entries\r\n", cnt0092);

            int cnt1000=0;
            for (p=(volatile char *)0x7c1000; p<(volatile char *)0x7c25cc; p+=5)
                if (((*p)&0x7f)!=0x7f)
                    cnt1000++;
            uprintf("1000 list contains %d entries\r\n", cnt1000);

            int vidx= 0;
            int cntvotes=0;
            for (p=(volatile char *)0x7d4000; p<(volatile char *)0x7da000; p+=8, vidx++)
                if (((*p)&0x7f)!=0x7f) {
                    if (cntvotes<20)
                        uprintf("%4d: %02d : %02d\r\n", vidx, p[0]&0x7f, p[1]&0x7f);
                    else if (cntvotes==32)
                        uprintf(" .... more votes, not listed\r\n");
                    cntvotes++;
                }
            uprintf("total %d votes\r\n", cntvotes);
        }
        else {
            usage();
        }
    }
}


#ifdef TEST
int main(int argc, char **argv)
{
    entrypoint();
    return 0;
}
#endif
