#ifdef TEST
#include <stdio.h>
#endif

#include <string.h>
#include "misc.h"
#include "display.h"
#include "io_funcs.h"
#include "uart.h"

unsigned char chesspiece_picture[16][12]= {
    {0x00, 0x00,  0x00, 0x00,  0x00, 0x00,  0x00, 0x00,  0x00, 0x00,  0x00, 0x00},
    {0x00, 0x00,  0x00, 0x00,  0x00, 0x00,  0x00, 0x00,  0x00, 0x00,  0x01, 0x80},
    {0x00, 0x00,  0x01, 0x00,  0x00, 0x80,  0x00, 0x00,  0x00, 0x00,  0x02, 0x40},
    {0x00, 0x00,  0x03, 0x80,  0x01, 0x40,  0x00, 0x00,  0x02, 0x20,  0x01, 0x80},
    {0x00, 0x00,  0x06, 0xe0,  0x01, 0x40,  0x0d, 0xb0,  0x05, 0x50,  0x1c, 0x38},
    {0x01, 0xc0,  0x04, 0x30,  0x02, 0xe0,  0x0a, 0x50,  0x12, 0x24,  0x36, 0x6c},
    {0x02, 0x20,  0x08, 0x10,  0x04, 0xb0,  0x08, 0x10,  0x2b, 0x6a,  0x23, 0xc4},
    {0x02, 0x20,  0x09, 0x88,  0x04, 0x90,  0x07, 0xe0,  0x1d, 0x5c,  0x21, 0x84},
    {0x01, 0x40,  0x12, 0x48,  0x04, 0x90,  0x04, 0x20,  0x09, 0xc8,  0x24, 0x24},
    {0x01, 0x40,  0x14, 0x88,  0x04, 0x10,  0x04, 0x20,  0x04, 0x10,  0x12, 0x48},
    {0x02, 0x20,  0x09, 0x08,  0x06, 0x20,  0x04, 0x20,  0x04, 0x10,  0x0a, 0x50},
    {0x02, 0x20,  0x03, 0xf8,  0x02, 0x60,  0x0f, 0xf0,  0x07, 0xf0,  0x0f, 0xf0},
    {0x04, 0x10,  0x04, 0x08,  0x3e, 0x7c,  0x18, 0x18,  0x0c, 0x18,  0x18, 0x18},
    {0x08, 0x08,  0x08, 0x04,  0x21, 0x84,  0x10, 0x08,  0x08, 0x08,  0x10, 0x08},
    {0x0f, 0xf8,  0x0f, 0xfc,  0x3e, 0x7c,  0x1f, 0xf8,  0x0f, 0xf8,  0x1f, 0xf8},
    {0x00, 0x00,  0x00, 0x00,  0x00, 0x00,  0x00, 0x00,  0x00, 0x00,  0x00, 0x00},
};



// width, height, row, col are all in bits.
void copy_5x8_from_bitmap(unsigned char *bitmap, position_t width, position_t height,
        position_t row, position_t col, charbitmap_t chardata)
{
    position_t bytewidth=((width+7)/8);
    bitmap+=(col/8)+bytewidth*row;
    position_t bitofs= col%8;
    static unsigned char mask1list[8]={0xf8, 0x7c, 0x3e, 0x1f, 0x0f, 0x07, 0x03, 0x01};
    static int shift1list[8]={  3,    2,    1,    0,   -1,   -2,   -3,   -4};
    static unsigned char mask2list[8]={0x00, 0x00, 0x00, 0x00, 0x80, 0xc0, 0xe0, 0xf0};
    static int shift2list[8]={  0,    0,    0,    0,    7,    6,    5,    4};
    unsigned char mask1= mask1list[bitofs];
    unsigned char mask2= mask2list[bitofs];
    int shift1= shift1list[bitofs];
    int shift2= shift2list[bitofs];

    position_t line;
    for (line=0 ; line<8 ; line++) {
#ifdef TESTDEBUG
        printf("bm_%08lx -> cd=%08lx  [%02x:%02x] %d,%d\n", bitmap, chardata, mask1, mask2, shift1, shift2);
#endif
        chardata[line]= (shift1<0? ((*bitmap&mask1)<<-shift1): ((*bitmap&mask1)>>shift1))
            | ((bitmap[1]&mask2)>>shift2);
#ifdef TESTDEBUG
        printf("bits=%08lx\n", chardata[line]);
#endif
        bitmap += bytewidth;
        watchdog();
    }
}

// (w,h) can be (2*4*5, 1*2*8)  or (1*4*5, 2*2*8)
//          (40, 16)  or (20, 32)
//
// line0==0 : 20 wide, 32 high
// line0==1 : 40 wide, 16 high
void copybitmap2display(
        unsigned char *bitmap, position_t width, position_t height, 
        position_t brow, position_t bcol, 
        position_t line0, position_t pos0)
{
    cgram_t cdata[2];
    ddram_t ddata[2];
    memset(cdata, 0, sizeof(cdata));
    memset(ddata, 0x20, sizeof(ddata));

    position_t charidx=0;

#ifdef TESTDEBUG
    printf("copy %d,%d -> %d,%d  cd=%8lx-%08lx\n", brow, bcol, pos0, line0, &cdata, &cdata[1][7][7]);
#endif
    position_t dispidx=0;
    position_t lineidx=line0;
    position_t r;
    for (r=brow ; r<height && r<brow+(line0?16:32); r+=8) {
        position_t posidx=pos0;
        position_t c;
        for (c=bcol ; c<width && c<bcol+(line0?40:20); c+=5) {
#ifdef TESTDEBUG
            printf("copy %d,%02d  to %d:(%d,%d)=%d  %08lx\n", r,c, dispidx, lineidx,posidx, charidx&7, &cdata[dispidx][charidx][0]);
#endif
            copy_5x8_from_bitmap(bitmap, width, height, r, c, cdata[dispidx][charidx]);
            ddata[dispidx][lineidx][posidx++] = (charidx++)&0x7;
            charidx &= 7;
            if (line0==0 && (charidx&3)==0)
                break;
        }
        //ddata[dispidx][lineidx][posidx++]=0x30+dispidx*4+lineidx;
        //ddata[dispidx][lineidx][posidx++]=0x30+charidx;

        lineidx++;
        if (lineidx==2) {
            dispidx++;
            lineidx=0;
        }
    }

    //load_cgram(4, 0, cdata[0], 64);
    //load_ddram(4, ddata[0], 128);

    dispinsn(8,1);
    dispinsn(0x10,1);
    load_cgram(8, 0, cdata[0], 64);
    load_cgram(0x10, 0, cdata[1], 64);
    if (line0==1) {
        load_ddram(8,    1, pos0, ddata[0][1]+pos0, 8);
        load_ddram(0x10, 0, pos0, ddata[1][0]+pos0, 8);
    }
    else {
        load_ddram(8,    0, pos0, ddata[0][0]+pos0, 4);
        load_ddram(8,    1, pos0, ddata[0][1]+pos0, 4);
        load_ddram(0x10, 0, pos0, ddata[1][0]+pos0, 4);
        load_ddram(0x10, 1, pos0, ddata[1][1]+pos0, 4);
    }

    wdelay(2000);
}

// bitmap, 8*5 wide, 8*4 high

typedef unsigned char chessboard_t[8*4][5];

char bm_blackfield[4]= {
    0xf8,   // *****xxx
    0xf8,   // *****xxx
    0xf8,   // *****xxx
    0xf8,   // *****xxx
};
char bm_whitefield[4]= {
    0x00,   // .....xxx
    0x00,   // .....xxx
    0x00,   // .....xxx
    0x00,   // .....xxx
};
char bm_pawn[3]= {
    0x00,   // ...xxxxx
    0x40,   // .*.xxxxx
    0x00,   // ...xxxxx
};
char bm_knight[3]= {
    0xe0,   // ***xxxxx
    0x80,   // *..xxxxx
    0x80,   // *..xxxxx
};
char bm_queen[3]= {
    0xa0,   // *.*xxxxx
    0x40,   // .*.xxxxx
    0xa0,   // *.*xxxxx
};
char bm_king[3]= {
    0x00,   // ...xxxxx
    0xa0,   // *.*xxxxx
    0xe0,   // ***xxxxx
};
char bm_rook[3]= {
    0x00,   // ...xxxxx
    0xe0,   // ***xxxxx
    0x00,   // ...xxxxx
};
char bm_bishop[3]= {
    0x20,   // ..*xxxxx
    0x40,   // .*.xxxxx
    0x80,   // *..xxxxx
};

typedef enum {
    CP_PAWN=0,
    CP_ROOK=1,
    CP_BISHOP=2,
    CP_KNIGHT=3,
    CP_QUEEN=4,
    CP_KING=5,
} chesspiece_t;

typedef enum {
    CP_BLACK=0,
    CP_WHITE=1,
} piececolor_t;

char *chesspieces[6]= {
    bm_pawn,
    bm_rook,
    bm_bishop,
    bm_knight,
    bm_queen,
    bm_king,
};
char bm_getpixel(
        char *bm, short w, short h,
        short row, short col)
{
    position_t bytewidth=((w+7)/8);
    return bm[bytewidth*row+(col/8)] & 0x80>>(col%8);
}
void bm_setpixel(
        char *bm, short w, short h,
        short row, short col, char value)
{
    position_t bytewidth=((w+7)/8);
    if (value)
        bm[bytewidth*row+(col/8)] |= (0x80>>(col%8));
    else
        bm[bytewidth*row+(col/8)] &= ~(0x80>>(col%8));
}
void copy_bm2bm( 
        char *srcbm, short srcw, short srch, 
        char *dstbm, short dstw, short dsth,
        short row, short col)
{
    short x,y;
    for (y=0 ; y<srcw ; y++)
        for (x=0 ; x<srcw ; x++)
            bm_setpixel(dstbm, dstw, dsth, row+y, col+x, bm_getpixel(srcbm, srcw, srch, y, x));
}

//  01234567
//0 .*.*.*.*   .=white field
//1 *.*.*.*.   *=black field
//2 .*.*.*.*
//3 *.*.*.*.
//4 .*.*.*.*
//5 *.*.*.*.
//6 .*.*.*.*
//7 *.*.*.*.

void clear_field(chessboard_t board, short x, short y)
{
    copy_bm2bm( ((x+y)&1)?bm_whitefield:bm_blackfield, 5,4,
            (char*)board, 8*5, 8*4, y*4, x*5 );
}
void erase_chessboard(chessboard_t board)
{
    short x,y;
    for (y=0 ; y<8 ; y++)
        for (x=0 ; x<8 ; x++)
            clear_field(board, x, y);
}
void draw_piece(chessboard_t board, short x, short y, chesspiece_t type, piececolor_t color)
{
    if (color) {
        // white
        copy_bm2bm(chesspieces[type], 3, 3, (char*)board, 8*5, 8*4, y*4, x*5+1);
    }
    else {
        // black
        char piece[3];
        piece[0]= 0xe0 ^ chesspieces[type][0];
        piece[1]= 0xe0 ^ chesspieces[type][1];
        piece[2]= 0xe0 ^ chesspieces[type][2];

        copy_bm2bm(piece, 3, 3, (char*)board, 8*5, 8*4, y*4, x*5+1);
    }
}
void init_chessboard(chessboard_t board)
{
    short x;
    erase_chessboard(board);
    for (x=0 ; x<8 ; x++) {
        draw_piece(board, x, 1, CP_PAWN, CP_BLACK);
        draw_piece(board, x, 6, CP_PAWN, CP_WHITE);
    }
    draw_piece(board, 0,0, CP_ROOK, CP_BLACK);
    draw_piece(board, 7,0, CP_ROOK, CP_BLACK);
    draw_piece(board, 0,7, CP_ROOK, CP_WHITE);
    draw_piece(board, 7,7, CP_ROOK, CP_WHITE);

    draw_piece(board, 1,0, CP_KNIGHT, CP_BLACK);
    draw_piece(board, 6,0, CP_KNIGHT, CP_BLACK);
    draw_piece(board, 1,7, CP_KNIGHT, CP_WHITE);
    draw_piece(board, 6,7, CP_KNIGHT, CP_WHITE);

    draw_piece(board, 2,0, CP_BISHOP, CP_BLACK);
    draw_piece(board, 5,0, CP_BISHOP, CP_BLACK);
    draw_piece(board, 2,7, CP_BISHOP, CP_WHITE);
    draw_piece(board, 5,7, CP_BISHOP, CP_WHITE);

    draw_piece(board, 3,0, CP_QUEEN, CP_BLACK);
    draw_piece(board, 3,7, CP_QUEEN, CP_WHITE);

    draw_piece(board, 4,0, CP_KING, CP_BLACK);
    draw_piece(board, 4,7, CP_KING, CP_WHITE);
}
void test_uart()
{
#ifndef TEST
    int i;
    char c;
    //char buf[32];
    load_ddram(4,0,5, "uart1",-1);
    init_uart(0);
    for (c='a' ; c<='z' ; c++)
        write_uart(0, c);
    load_ddram(4,0,11, "uart2",-1);
    init_uart(1);
    for (c='z' ; c>='a' ; c--)
        write_uart(1, c);
    for (i=0 ; i<4 ; i++) {
        c=read_uart(1);
        write_uart(1, c^0x20);
        write_uart(1, c^0x20);
    }
#endif
}
#ifdef TEST
void print_bitmap(char *bm, short width, short height)
{
    short r,c;
    position_t bytewidth=((width+7)/8);
    for(r=0 ; r<height ; r++) {
        for (c=0 ; c<width/8 ; c++)
            printf(" %c%c%c%c%c%c%c%c",
                    bm[r*bytewidth+c]&0x80?'*':'.',
                    bm[r*bytewidth+c]&0x40?'*':'.',
                    bm[r*bytewidth+c]&0x20?'*':'.',
                    bm[r*bytewidth+c]&0x10?'*':'.',
                    bm[r*bytewidth+c]&0x08?'*':'.',
                    bm[r*bytewidth+c]&0x04?'*':'.',
                    bm[r*bytewidth+c]&0x02?'*':'.',
                    bm[r*bytewidth+c]&0x01?'*':'.');
        printf("\n");
        if ((r&3)==3)
            printf("\n");
    }
}
#endif

#ifdef TEST
void entrypoint()
#else
int main()
#endif
{
    position_t line0= 0;
    int chr;
    int id;
    int chr0='0';

    wbeep(500);
    init_nedap();
    wdelay(1000);

    init_display(4);
    load_ddram(4,0,0, "init",-1);
    wbeep(1000);

    backlight(1);

    watchdog();
    test_uart();

    for (id=4 ; id<=0x10 ; id*=2) {
        init_display(id);
        dispinsn(id, 0x80);
        for (chr=chr0 ; chr<chr0+40 ; chr++) {
            dispdata(id, chr);
            watchdog();
        }
        chr0++;
        dispinsn(id, 0x80+0x40);
        for (chr=chr0 ; chr<chr0+40 ; chr++) {
            dispdata(id, chr);
            watchdog();
        }
        chr0++;
    }
    delay(20000);

    init_display(8);
    init_display(0x10);
    chessboard_t board;
    init_chessboard(board);
#ifdef TEST
    print_bitmap((char*)board, 40, 32);
#endif
    while(1) {
        copybitmap2display(
                (unsigned char*)&board, 40, 32, 
                0, 0, 
                line0, 0);
        copybitmap2display(
                (unsigned char*)&board, 40, 32, 
                0, line0?40:20, 
                line0, line0?8:4);
#ifdef TEST
        return;
#endif
    }
}


#ifdef TEST
int main(int argc, char **argv)
{
    entrypoint();
    return 0;
}
#endif
