/*
 *	MAIN.C
 *	Tom Kerrigan's Simple Chess Program (TSCP)
 *
 *	Copyright 1997 Tom Kerrigan
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "defs.h"
#include "data.h"
#include "protos.h"

#ifndef TEST
#include <compat.h>
#include "uart.h"
#include "display.h"
#include "keyboard.h"

char debug[256];
ddramline_t displaystring;
#define printf( args... ) write( 1, debug, sprintf( debug, args ))

#define l1_printf( args... ) \
{ memset(displaystring+sprintf((char*)displaystring, args),' ',40); load_ddram( 8, 0, 0, displaystring, 21); }
#define l2_printf( args... ) \
{ memset(displaystring+sprintf((char*)displaystring, args),' ',40); load_ddram( 8, 1, 0, displaystring, 21); }
#define l3_printf( args... ) \
{ memset(displaystring+sprintf((char*)displaystring, args),' ',40); load_ddram(16, 0, 0, displaystring, 21); }
#define l4_printf( args... ) \
{ memset(displaystring+sprintf((char*)displaystring, args),' ',40); load_ddram(16, 1, 0, displaystring, 21); }

void CLS() { memset(displaystring,' ',40); load_ddram( 8, 0, 0, displaystring, 40); load_ddram( 8, 1, 0, displaystring, 40);
load_ddram(16, 0, 0, displaystring, 40); load_ddram(16, 1, 0, displaystring, 40); }

int get_ms()
{
        return get_time() * 10;
}
#else
int get_ms()
{
        return time(NULL) * 1000;
}
#endif

#define KEYCODE_NEWGAME 0x1d0b
#define KEYCODE_COMPTURN 0x1d0e
#define KEYCODE_ENTER 0x2223
#define KEYCODE_UNDO 0x1d1d

char y_map[] = "8XX7XX6XX5XX4XX3XX2XX1";

void which_command( int key_code, char *buffer, int *bufidx ) {
  char x, y;

  switch( key_code ) {
    case KEYCODE_NEWGAME:
    case KEYCODE_COMPTURN:
      memcpy( buffer, key_code==KEYCODE_NEWGAME?"new\0":"on\0\0", 4 );
      *bufidx=-1;
      return;
    case KEYCODE_ENTER:
      if(*bufidx<4)
        l4_printf( "%s ... incomplete", buffer)
      else {
        l4_printf(" ");
        *bufidx=-1;
      }
      return;
    case KEYCODE_UNDO:
      if(*bufidx==0) {
        l4_printf("... cant undo");
      } else {
        *bufidx-=2; if(*bufidx<0) *bufidx=0; buffer[*bufidx]=0;
        l4_printf("%s", buffer);
      }
      return;
    default:
      break;
  }

#define xp ((key_code>>8)&255)
#define yp ((key_code)&255)

  if( (xp < 9) || (xp>0x17) || !(xp&1)) goto beep;
  x = 'a' + ((xp - 9) >> 1);

  if( (yp < 6) || (yp>0x1b)) goto beep;
  y = y_map[yp - 6];
  if ( y == 'X' ) goto beep;

#undef xp
#undef yp

  if ( *bufidx == 4 ) {
    l4_printf("%s ... complete", buffer);
    return;
  }

  buffer[(*bufidx)++] = x;
  buffer[(*bufidx)++] = y;
  buffer[*bufidx    ] = 0;

  l4_printf( "%s", buffer);
  return;

beep:
  beep(100);
  l4_printf("%s ... incorrect",buffer);
  return;
}

// This is our state machine we try to interpret
// user interaction with
void fetch_command( char *buffer, int buflen ) {
  char c;
  int bufidx = 0, button_down = -2, key_code;
  
  print_board();
  l3_printf("Your Move:");
  l4_printf(" "); *buffer=0;

  while(1) {
    if(!read_uart_noblock(1, &c)) {
      if(bufidx+1>=buflen) {
        buffer[buflen-1]=0;
        return;
      }
      buffer[bufidx++]=c;
      if(c=='\n') goto done;
    }
    key_code = read_from_keyboard();
    if( key_code == -2 ) { // is this a release?
      if (button_down >= 0) {
        which_command( button_down, buffer, &bufidx);
        button_down = -2;
        if( bufidx < 0 ) // command finished?
          goto done;
      }
    } else
      if( key_code >= 0)
        button_down = key_code;
  }

done:
  l3_printf(" ");
}


/* main() is basically an infinite loop that either calls
   think() when it's the computer's turn to move or prompts
   the user for a command (and deciphers it). */
int computer_side;

int main()
{
	char s[256];
	int m;

	printf("\r\n");
	printf("Tom Kerrigan's Simple Chess Program (TSCP)\r\n");
	printf("version 1.81, 2/5/03\r\n");
	printf("Copyright 1997 Tom Kerrigan\r\n");
	printf("\r\n");
	printf("\"help\" displays a list of commands.\r\n");
	printf("\r\n");
	init_hash();
	init_board();
        CLS();
        print_board();
	gen();
	computer_side = DARK;
	max_time = 10000;
	max_depth = 4;
	for (;;) {
		if (side == computer_side) {  /* computer's turn */
			
			/* think about the move and make it */
                        print_board();
                        l1_printf("Thinking...");
			think(1);
			if (!pv[0][0].u) {
				printf("(no legal moves)\r\n");
				l1_printf("(No legal moves)");
				computer_side = EMPTY;
				continue;
			}
			printf("Computer moves: %s\r\n", move_str(pv[0][0].b));
			l1_printf("Computer moves: %s", move_str(pv[0][0].b));

			makemove(pv[0][0].b);
			ply = 0;
			gen();
			print_result();
			continue;
		}

		/* get user input */
		printf("tscp> ");
#ifndef TEST
		fetch_command(s,256);
#else
                scanf("%s", s);
#endif
        	open_book();
		if (!strcmp(s, "on")) {
			computer_side = side;
			continue;
		}
		if (!strcmp(s, "off")) {
			computer_side = EMPTY;
			continue;
		}
		if (!strcmp(s, "st")) {
			scanf("%d", &max_time);
			max_time *= 1000;
			max_depth = 32;
			continue;
		}
		if (!strcmp(s, "sd")) {
			scanf("%d", &max_depth);
			max_time = 1 << 25;
			continue;
		}
		if (!strcmp(s, "undo")) {
			if (!hply)
				continue;
			computer_side = EMPTY;
			takeback();
			ply = 0;
			gen();
			continue;
		}
		if (!strcmp(s, "new")) {
			computer_side = DARK;
			init_board();
 			CLS();
			gen();
			continue;
		}
		if (!strcmp(s, "d")) {
			print_board();
			continue;
		}
		if (!strcmp(s, "bye")) {
			printf("Share and enjoy!\r\n");
			break;
		}
		if (!strcmp(s, "help")) {
			printf("on - computer plays for the side to move\r\n");
			printf("off - computer stops playing\r\n");
			printf("st n - search for n seconds per move\r\n");
			printf("sd n - search n ply per move\r\n");
			printf("undo - takes back a move\r\n");
			printf("new - starts a new game\r\n");
			printf("d - display the board\r\n");
			printf("bench - run the built-in benchmark\r\n");
			printf("bye - exit the program\r\n");
			printf("xboard - switch to XBoard mode\r\n");
			printf("Enter moves in coordinate notation, e.g., e2e4, e7e8Q\r\n");
			continue;
		}

		/* maybe the user entered a move? */
		m = parse_move(s);
		if (m == -1 || !makemove(gen_dat[m].m.b)) {
			printf("Illegal move.\r\n");
			l4_printf("Illegal move.");
                } else {
			ply = 0;
			gen();
			print_result();
		}
	}
	close_book();
	return 0;
}


/* parse the move s (in coordinate notation) and return the move's
   index in gen_dat, or -1 if the move is illegal */

int parse_move(char *s)
{
	int from, to, i;

	/* make sure the string looks like a move */
	if (s[0] < 'a' || s[0] > 'h' ||
			s[1] < '0' || s[1] > '9' ||
			s[2] < 'a' || s[2] > 'h' ||
			s[3] < '0' || s[3] > '9')
		return -1;

	from = s[0] - 'a';
	from += 8 * (8 - (s[1] - '0'));
	to = s[2] - 'a';
	to += 8 * (8 - (s[3] - '0'));

	for (i = 0; i < first_move[1]; ++i)
		if (gen_dat[i].m.b.from == from && gen_dat[i].m.b.to == to) {

			/* if the move is a promotion, handle the promotion piece;
			   assume that the promotion moves occur consecutively in
			   gen_dat. */
			if (gen_dat[i].m.b.bits & 32)
				switch (s[4]) {
					case 'N':
						return i;
					case 'B':
						return i + 1;
					case 'R':
						return i + 2;
					default:  /* assume it's a queen */
						return i + 3;
				}
			return i;
		}

	/* didn't find the move */
	return -1;
}


/* move_str returns a string with move m in coordinate notation */

char *move_str(move_bytes m)
{
	static char str[6];

	char c;

	if (m.bits & 32) {
		switch (m.promote) {
			case KNIGHT:
				c = 'n';
				break;
			case BISHOP:
				c = 'b';
				break;
			case ROOK:
				c = 'r';
				break;
			default:
				c = 'q';
				break;
		}
		sprintf(str, "%c%d%c%d%c",
				COL(m.from) + 'a',
				8 - ROW(m.from),
				COL(m.to) + 'a',
				8 - ROW(m.to),
				c);
	}
	else
		sprintf(str, "%c%d%c%d",
				COL(m.from) + 'a',
				8 - ROW(m.from),
				COL(m.to) + 'a',
				8 - ROW(m.to));
	return str;
}


/* print_board() prints the board */
inline char which_letter(i) {
  if(color[i]==EMPTY) return (i^i>>3)&1? '.':' ';
  return piece_char[piece[i]]+(color[i]==LIGHT?0:('a' - 'A'));
}

void print_board()
{
        char board[18];
	int i, j;
	
	printf("\r\n8 ");
	for (i = 0; i < 64; ++i) {
		switch (color[i]) {
			case EMPTY:
				printf(" .");
				break;
			case LIGHT:
				printf(" %c", piece_char[piece[i]]);
				break;
			case DARK:
				printf(" %c", piece_char[piece[i]] + ('a' - 'A'));
				break;
		}
		if ((i + 1) % 8 == 0 && i != 63)
			printf("\r\n%d ", 7 - ROW(i));
	}
	printf("\r\n\r\n   a b c d e f g h\r\n\r\n");

        backlight(1);
        board[8]=board[16]=' ';
        for (i = 0; i < 4; ++i) {
          for(j = 0; j < 8; ++j) {
            board[j  ]=which_letter(8*i+j);
            board[j+9]=which_letter(8*i+j+32);
          }
          load_ddram(8+(i>>1)*8, i&1, 22, board, 17);
        }
}


/* print_result() checks to see if the game is over, and if so,
   prints the result. */

void print_result()
{
	int i;

	/* is there a legal move? */
	for (i = 0; i < first_move[1]; ++i)
		if (makemove(gen_dat[i].m.b)) {
			takeback();
			break;
		}
	if (i == first_move[1]) {
		if (in_check(side)) {
			if (side == LIGHT) {
				printf("0-1 {Black mates}\r\n"); 
			} else {
				printf("1-0 {White mates}\r\n");
			}
			l1_printf( "%s lose", ((side == computer_side) ? "I " : "You "));
		}
		else {
			printf("1/2-1/2 {Stalemate}\r\n");
			l1_printf( "Stalemate" );
		}
	}
	else if (reps() == 3)
		printf("1/2-1/2 {Draw by repetition}\r\n");
	else if (fifty >= 100)
		printf("1/2-1/2 {Draw by fifty move rule}\r\n");
}
